{Feedback Servo Driver}

CON
        _clkmode = xtal1 + pll16x    'Standard clock mode 
        _xinfreq = 5_000_000         '* crystal frequency = 80 MHz                    
        dutyScale = 360, dcmin = 33, dcrange = 961  'feedback scaling constants
        gain = 10                    'speed proportional to error
        
VAR                                       
  long  MyParm[10]                'passing addresses, etc to pasm cog
  byte  Cog
  long  PulseWidth                 'control pulse width in clock tics
  long  tCycle, dc, theta          'feedback variables
  long  PeriodValid                '
  long  error                      'position error  
  long  OneUSec                    'clock tics in one microsecond
  
PUB start(CPin, FPin)                   'start the control cog
  OneUsec := clkfreq/1_000_000           'constant
  MyParm[0] := @PulseWidth               'address of pulse width to motor
  MyParm[1] := @dc                       'feedback
  MyParm[2] := clkfreq / 250              'tics in 1/250th second (4 msec)               
  MyParm[3] := CPin                      'control pin number
  MyParm[4] := 1 << FPin                 'feedback pin mask 
  cog := cognew(@BCog, @MyParm)           'start up the pasm cog
  dira[26] := 1                           'instrumentation
  dira[27] := 1
  return cog            

PUB SetPos(degrees)                 'come here with desired position
    repeat until (||(degrees - getpos)) < 2   '+-1 degree was as close as I could get
      error := (degrees - getpos)    'signed error
      if (error > 0)
         outa[26] := 1                'positive:  go cw   
         error <#= 5                 'limit to avoid extreme pulse width
         width(1500 - (gain * error)) 'set width proportional to error
      else
         outa[27] := 1 
         error := || error             'negative: go ccw
         error <#= 5
         width(1500 + (gain * error))    
    outa[26..27] := 0                  'turn off indicators         
    width(1500)                        'stop servo                           
    return getpos                      'return actual position in degrees

PUB GetPos
    return ((dc-dcmin) * 359) / dcrange   'scale to degrees 
    
PUB Width(Usec)                              'set pulse widt
    pulsewidth := USec * OneUSec              'scale to clock tics
    
DAT   '***********pasm cog******************
BCog    org   0                'generates control pulse, returns DC in range 33..994               '
        mov AdPar,Par          'get the hub address of control PulseWidth
        rdlong AdPWidth, AdPar '
        add AdPar, #4           'to address of duty cycle
        rdlong AdDC, AdPar  
        add AdPar, #4             'to constant clk tics in 4 msec
        rdlong Hz250, AdPar
        add AdPar, #4             'to control pin number
        rdlong ContPin, AdPar
        add AdPar, #4              'to feedback pin mask
        rdlong FBPinMask, AdPar
        mov work1, #%00100          'pina follows phsa[31]
        shl work1, #26               'to control field
        mov ctra, work1             'start up counter a
        movs ctra, ContPin          'control pin
        mov frqa, #1                'count by one
        mov CPinMask, #1             'form mask with single "one"
        shl CPinMask, ContPin        'scale
        mov dira, CPinMask           'enable output
        or dira, RFBMask            'engineering
        mov timer, cnt              'initialize 250 hz timer
        add timer, Hz250                  
loop    rdlong work1, AdPWidth      'get control pulse width
        neg phsa, work1             'start a pulse
        mov Count5, #5              'will return five feedback pairs
l1      or outa, RFBMask            'scope
        waitpne FBPinMask, FBPinMask 'wait for feedback low
        waitpeq FBPinMask, FBPinMask 'then high
        mov work1, cnt               'beginning of high feedback pulse
        waitpne FBPinMask, FBPinMask 'wait for end of high pulse
        mov HiPer, cnt
        waitpeq FBPinMask, FBPinMask 'now wait for end of low pulse
        mov LowPer, cnt
        sub LowPer, HiPer           'calculate low period
        sub HiPer, work1            'and high period
        mov y, HiPer                'calculate total period
        add y, LowPer               '
        shr y, #1                   'keep within 16 bits
        andn y, Hi16                'just be certain
        mov x, HiPer                'set up dividend
        shl x, #9                   'scale by 1024
        call #divide
        andn x, Hi16                'discard remainder
        wrlong x, AdDC              'return dc
        andn outa, RFBMask 
        waitcnt timer, Hz250         '4 msec period
        djnz Count5, #L1             'five times (total of 20 msec)
        jmp #loop                   'forever

' Divide x[31..0] by y[15..0] (y[16] must be 0)
' on exit, quotient is in x[15..0] and remainder is in x[31..16]
'
divide  shl  y,#15    'get divisor into y[30..15]
        mov  t,#16    'ready for 16 quotient bits
:loop   cmpsub x,y wc 'y =< x? Subtract it, quotient bit in c
        rcl    x,#1    'rotate c into quotient, shift dividend
        djnz   t,#:loop 'loop until done
divide_ret ret          'quotient in x[15..0], 'remainder in x[31..16]         
                    
zero        long  0
RFBMask     long  %0010         'I/O 1
Hi16        long  $FFFF_0000    'high word
Hz250       res                 'clock tics in 4 msec  
AdPar       res                 'parameter list                                                     
AdPWidth    res                 'address of control pulse width in hub
AdDC        res                 'address of duty cycle feedback
HiPer       res                 'high period in clks
LowPer      res                 'low period
ContPin     res                 'control pin number
FBPinMask   res                 'feedback pin mask
timer       res                 'clock at end of next period
work1       res                 'generally useful
CPinMask    res                 'control pin mask
Count5      res                 'update position 5 times per 20 msec
x           res                 'pasm divide variables
y           res
t           res          